/**************************************************************

File Name:	Fftrad4_simd.asm	ADSP-21364 Radix-4 Complex Fast Fourier Transform 

Author:     	Karl Schwarz & Raimund Meyer, Universitaet Erlangen Nuernberg       
Version:    	25-APR-91, Ronnin Yee, Analog Devices, DSP div.
		
Date Modified:	05/22/00	GJO		Modified for ADSP-21364 SIMD Multichannel
				05/25/00	GJO		Code verified on rev 0.1 Silicon
				09/19/04	JDT		Ported to VDSP++ 3.5		

Description:
	ADSP-21364 multichannel SIMD routine performs mulitchannel complex, radix 4 Fast 
	Fourier Transforms (FFT) on 2 data samples. 
	The FFT length (N) must be a power of 4 and a minimum of 64 points.
	The real parts of the input data samples are placed in DM and the complex parts in PM. This data is 
	destroyed during the course of the computation. The real outputs of
	the FFT is placed in DM and the complex is placed in PM.

	Since this routine takes care of all necessary address digit-reversals, the 
	input and output data are in normal order. The digit reversal is accomplished
	by using a modified radix 4 butterfly throughout which swaps the inner two
	nodes resulting with bit reversed data. The digit reversal is completed by
	bit reversing the data in the final stage.
	
	To implement an inverse FFT, you only have to (1) swap the incoming datas real
	and imaginary parts, (2) run the forward FFT, (3) swap the outgoing datas
	real and imaginary parts and (4) scale the data by 1/N.

	This radix-4 FFT routine requires two twiddle factor tables stored in memory.
	Each table consist of 3N/4 factors.  Twiddle factors, Wn, are the coefficients 
	used to combine results from a previous stage to form inputs to the next stage.
	A "C" language utility for generating the two tables is called "twidrad4.exe".
	The twiddle factors are read into explicit and implicit registers using 
	broadcasting.

Calling Parameters:
		    costwid table at DM : cosine    length 3*N/4
    		sintwid table at PM : sine      length 3*N/4
		    real  inputs  at DM : redata     length 2*N, normal order
		    imag  inputs  at PM : imdata     length 2*N, normal order

Assumptions:
			Real and imaginary input arrays are interleaved for multichannel
			mode of operation.
			Imaginary input, imaginary output, instructions, and sine twiddle 
			factors are stored in Block 0.
			Real input, real output, and cosine twiddle factors are stored in Block 1.

			Block 0
			Imaginary Input = {Ai[0],Bi[0],Ai[1],Bi[1],... }
			Sine Twiddle = { Ws[0],Ws[1],Ws[2],... }
			Imaginary Output = {Ai'[0],Bi'[0],Ai'[1],Bi'[1],... }
			
			Block 1
			Real Input = {Ar[0],Br[0],Ar[1],Br[1],... }
			Cosine Twiddle = { Wc[0],Wc[1],Wc[2],... }
			Real Output = {Ar'[0],Br'[0],Ar'[1],Br'[1],... }
			
Return Values:
			multichannel real  output at DM : refft      length 2*N, normal order
		    multichannel imag  output at DM : imfft      length 2*N, normal order

		    (Note:  Because the bit reversed addressing mode is used with the arrays
		    refft and imfft, they must start at addresses that are integer
		    multiples of the length (N) of the transform, (i.e. 0,N,2N,3N,...).
		    This is accomplished by specifying two segments starting at those addresses
		    in the architecture file and placing the variables alone in their 
		    respective segments. These addresses must also be reflected in the 
		    preprocessor variables ORE and OIM in bit reversed format.)

Registers Affected:
		    All I, M, L and R registers.
		    Three levels of looping.

Cycle Count:
		    FFT Length   cycles   ms @ 100 MHz CLK SISD	  ms @ 100 MHz CLK SIMD
		    ----------   ------   ---------------------   ------------------------
				64          832            0.008                   0.004  
		       256         3794        	   0.038                   0.019
		      1024        18228       	   0.182  				   0.091                   
    		  4096        86615     	   0.866				   0.433                   
		     16384       403065      	   4.030				   2.015                   

		    First Stage  -   8 cycles per radix-4 butterfly
		    Other Stages -  14 cycles per radix-4 butterfly

Memory Usage:
		    pm code = 191 words, pm data = 4.75*N words, dm data = 4.75*N words

Notes: 
			The constants below must be changed for different length FFTs
		    N       Number of points in FFT.  Must be a power of four, minimum of 64.
		    STAGES  Set to log4(N) or (log(N)/log(4))
		    OST     = bitrev(32 bit N/2)
		    ORE     = bitrev(32 bit addr of output real in dm), addr is 0,N,2N,3N,...
		    OIM     = bitrev(32 bit addr of output imag. in dm), addr is 0,N,2N,3N,...

			Circular Buffering Issue:
			Because SIMD or Long word access transfer two 32-bit words, programs
			must be careful when using these accesses in circular buffering.
			It is important that SIMD or Long word accesses do not cross a circular
		 	buffer boundary.  If a SIMD	mode access occurs using a circular buffer
			index register that points to the last location in the circular buffer
		 	(end of buffer), the resulting access transfers the last location in
		 	the circular buffer and the	first location outside the buffer (end of buffer + 1).

************************************************************************************/

/***********************************************************************************/
#define N       1024
#define STAGES  5   /* 64=3, 256=4 , 1024=5, 4096=6 , 16384=7 */			
#define OST     0x00200000		
#define ORE     0x000dd000  	/* DM: ORE=0x000bb000, bitrev to get the index */
#define OIM     0x00079000	  	/* PM: OIM=0x0009e000, bitrev to get the index */

/* include for symbolic definition of system regster bits */
#include "def21364.h"

.section/DM             seg_dmda;
.VAR    redata[2*N]="inreal1024.dat"; 	/* Input real data */
.VAR    cosine[3*N/4]="costw1024.dat";	/* Cosine twiddle factors */
.GLOBAL redata;


.section/DM             seg_dmre; /* this segment is an integer multiple of N */
.VAR    refft[2*N];               /* Output real fft data */
.GLOBAL refft;


.section/PM             seg_pmda;
.VAR    imdata[2*N]="inimag1024.dat"; /* Input imaginary data */
.VAR    sine[3*N/4]="sintw1024.dat";  /* Sine twiddle factors */
.GLOBAL imdata;

.section/PM             seg_pmim; /* this segment is an integer multiple of N */
.VAR    imfft[2*N];               /* Output imaginary fft data */
.GLOBAL imfft;




.section/PM             seg_rth; 	/* program starts at the reset vector */
reserved:	nop; nop; nop; nop;
reset:      idle; call fft; nop; nop;idle;



.section/PM             seg_pmco;
fft:
/*_______first stage radix-4 butterfly without twiddles______*/
			bit set mode1 CBUFEN | PEYEN;	/* enable circular buffering and SIMD */
			b0=redata;
			b1=redata+2*N/4;
			b2=redata+2*N/2;
			b3=redata+2*3*N/4;  
			i4=i0;
			i5=i1;
			i6=i2;
			i7=i3;
			m0=2;
			m8=m0;
			b8=imdata;
			b9=imdata+2*N/4;
			b10=imdata+2*N/2;
			b11=imdata+2*3*N/4;  
			i12=i8;
			i13=i9;
			i14=i10;
			i15=i11;
			l0 = 0;
			l1 = l0;
			l2 = l0;
			l3 = 2*N;     /* circular to prevent pointer overflow */
			l4 = l0;
			l5 = l0;
			l6 = l0;
			l7 = l0;
			l8 = l0;
			l9 = l0;
			l10 = l0;
			l11 = 2*N;    /* circular to prevent pointer overflow */
			l12 = l0;
			l13 = l0;
			l14 = l0;
			l15 = l0;
						f0=dm(i0,m0),   f1=pm(i8,m8);
						f2=dm(i2,m0),   f3=pm(i10,m8);
		f0=f0+f2,       f2=f0-f2,       f4=dm(i1,m0),   f5=pm(i9,m8);
		f1=f1+f3,       f3=f1-f3,       f6=dm(i3,m0),   f7=pm(i11,m8);
		f4=f6+f4,       f6=f6-f4;
		f5=f5+f7,       f7=f5-f7;
		f8=f0+f4,       f9=f0-f4;
		f10=f1+f5,      f11=f1-f5;

lcntr=N/4,      do fstage until lce;    /* do N/4 simple radix-4 butterflies */
		f12=f2+f7,      f13=f2-f7,      f0=dm(i0,m0),   f1=pm(i8,m8);
		f14=f3+f6,      f15=f3-f6,      f2=dm(i2,m0),   f3=pm(i10,m8);
		f0=f0+f2,       f2=f0-f2,       f4=dm(i1,m0),   f5=pm(i9,m8);
		f1=f1+f3,       f3=f1-f3,       f6=dm(i3,m0),   f7=pm(i11,m8);
		f4=f6+f4,       f6=f6-f4,       dm(i4,m0)=f8,   pm(i12,m8)=f10;
		f5=f5+f7,       f7=f5-f7,       dm(i5,m0)=f9,   pm(i13,m8)=f11;
		f8=f0+f4,       f9=f0-f4,       dm(i6,m0)=f12,  pm(i14,m8)=f14;
fstage:
		f10=f1+f5,      f11=f1-f5,      dm(i7,m0)=f13,  pm(i15,m8)=f15;

		l3=l0;      /* pointer overflow only problem in 1st stage */
		l11=l0;     /* pointer overflow only problem in 1st stage */

/*_____________Middle stages with radix-4 main butterfly___________________*/
			bit set mode1 BDCST1 | BDCST9;	/* enable broadcasting */
/* m0=2 and m8=2 is still preset */
			m1=-2;          /* reverse step for twiddles */
			m9=m1;  
			m2=3;           /* forward step for twiddles */
			m10=m2;
			m5=4;           /* first there are 4 groups */
			m7=1;			/* forward step for twiddles */
			m14=m7;
			r2=N/16;        /* with N/16 bflies in each group*/
			r3=2*N/16*3;    /* step to next group */

lcntr=STAGES-2, do mstage until lce;    /* do STAGES-2 stages */
			i1=cosine;      /* first real twiddle */
			i9=sine;        /* first imag twiddle */
			r8=redata;
			r9=imdata;
			r2=ashift r2 by 1;		/* r2x2 for SIMD index */
		 	i0=r8;          /* upper real path */
	r10=r8+r2,      i8=r9;          /* upper imaginary path */
			i7=r10;         /* second real input path */
	r10=r10+r2,     i4=r10;         /* second real output path */
			i2=r10;         /* third real input path */
	r10=r10+r2,     i5=r10;         /* third real output path */
			i3=r10;         /* fourth real input path */
	r10=r9+r2,      i6=r10;         /* fourth real output path */
			i15=r10;        /* second imag input path */
	r10=r10+r2,     i12=r10;        /* second imag output path */
			i10=r10;        /* third imag input path */
	r10=r10+r2,     i13=r10;        /* third imag output path */
			i11=r10;        /* fourth imag input path */
			i14=r10;        /* fourth imag output path */ 
			m4=r3;			
	r4=r3+1,	m12=r3;
			r2=ashift r2 by -1;  /* reverse r2: r2/2 */
	r4=r4+1,	m6=r2;
			m3=r4;
	r2=r2-1,        m11=r4;
			m13=r2;
	
lcntr=m5,       do mgroup until lce;    /* do m5 groups */
												f0=dm(i1,m7),   f5=pm(i15,m8);
f8=f0*f5,                                       f4=dm(i7,m0),   f1=pm(i9,m14);
f9=f0*f4;
f12=f1*f5,                                      f0=dm(i1,m7),   f5=pm(i11,m8);
f13=f1*f4,      f12=f9+f12,                     f4=dm(i3,m0),   f1=pm(i9,m14);
f8=f0*f4,                       f2=f8-f13;
f13=f1*f5;
f9=f0*f5,       f8=f8+f13,                      f0=dm(i1,m1),   f5=pm(i10,m8);
f13=f1*f4,      f12=f8+f12,     f14=f8-f12,     f4=dm(i2,m0),   f1=pm(i9,m9);
f11=f0*f4;
f13=f1*f5,                      f6=f9-f13;
f9=f0*f5,       f13=f11+f13,                    f11=dm(i0,0);
f13=f1*f4,      f8=f11+f13,     f10=f11-f13;
	   /*___________Do m13 radix-4 butterflies___________*/
lcntr=m13,       do mr4bfly until lce;
								f13=f9-f13,     f4=dm(i7,m0),   f5=pm(i15,m8);
		f2=f2+f6,       		f15=f2-f6,      f0=dm(i1,m7),   f1=pm(i9,m14);
f8=f0*f4,       f3=f8+f12,      f7=f8-f12,                      f9=pm(i8,0);
f12=f1*f5,      f9=f9+f13,      f11=f9-f13,     f13=f2;
f8=f0*f5,       f12=f8+f12,                     f0=dm(i1,m7),   f5=pm(i11,m8);
f13=f1*f4,      f9=f9+f13,      f6=f9-f13,      f4=dm(i3,m0),   f1=pm(i9,m14);
f8=f0*f4,                       f2=f8-f13,      dm(i0,m0)=f3,   pm(i8,m8)=f9;
f13=f1*f5,      f11=f11+f14,    f7=f11-f14,     dm(i4,m0)=f7,   pm(i12,m8)=f6;
f9=f0*f5,       f8=f8+f13,                      f0=dm(i1,m1),   f5=pm(i10,m8);
f13=f1*f4,      f12=f8+f12,     f14=f8-f12,     f4=dm(i2,m0),   f1=pm(i9,m9);
f11=f0*f4,      f3=f10+f15,     f8=f10-f15,                     pm(i13,m8)=f11;
f13=f1*f5,                      f6=f9-f13,      dm(i6,m0)=f8,   pm(i14,m8)=f7;
f9=f0*f5,       f13=f11+f13,                    f11=dm(i0,0);
mr4bfly:
f13=f1*f4,      f8=f11+f13,     f10=f11-f13,    dm(i5,m0)=f3;
	   /*___________End radix-4 butterfly_____________*/
/*                       dummy for address update      *               * */
						f13=f9-f13,     f0=dm(i1,m2),  f1=pm(i9,m10);
		f2=f2+f6,       f15=f2-f6,      f0=dm(i7,m4),  f1=pm(i15,m12);
		f3=f8+f12,      f7=f8-f12,                     f9=pm(i8,0);
		f9=f9+f13,      f11=f9-f13,     f0=dm(i2,m4);
		f9=f9+f2,       f6=f9-f2,       f0=dm(i3,m4),  f1=pm(i10,m12);
										dm(i0,m3)=f3,  pm(i8,m11)=f9;
		f11=f11+f14,    f7=f11-f14,     dm(i4,m3)=f7,  pm(i12,m11)=f6;
		f3=f10+f15,     f8=f10-f15,                    pm(i13,m11)=f11;
										dm(i6,m3)=f8,  pm(i14,m11)=f7;
mgroup:                                 dm(i5,m3)=f3,  f1=pm(i11,m12);

				r3=m4;
				r1=m5;
				r2=m6;
	r3=ashift r3 by -2;             /* groupstep/4 */
	r1=ashift r1 by 2;              /* groups x 4 */
				m5=r1;
mstage: r2=ashift r2 by -2;         /* butterflies/4 */
	
/*____________________Last radix-4 stage__________________________________*/
/* Includes bitreversal of the real and imaginary data in dm */

	bit set mode1 BR0 | BR8;              /* bitreversal in i0 and i8 */
/* with: m0=m8=2 preset */
	m1=8;							/* forward step for twiddles */
	m9=m1;
	i4=redata;                      /* real input */
	i7=redata+2*1;
	i2=redata+2*2;
	i3=redata+2*3;
	b0=ORE;  		/* real output array: integer multiple of (N) */
	m2=OST;			/* bit reversed modifer */
	i1=cosine;
	i12=imdata;                      /* image input */
	i15=imdata+2*1;
	i10=imdata+2*2;
	i11=imdata+2*3;
	b8=OIM;  		/* imaginary output array: integer multiple of (N) */
	m10=OST; 
	i9=sine;
	
												f0=dm(i1,m7),   f5=pm(i15,m9);
f8=f0*f5,                                       f4=dm(i7,m1),   f1=pm(i9,m14);
f9=f0*f4;
f12=f1*f5,                                      f0=dm(i1,m7),   f5=pm(i11,m9);
f13=f1*f4,      f12=f9+f12,                     f4=dm(i3,m1),   f1=pm(i9,m14);
f8=f0*f4,                       f2=f8-f13;
f13=f1*f5;
f9=f0*f5,       f8=f8+f13,                      f0=dm(i1,m7),   f5=pm(i10,m9);
f13=f1*f4,      f12=f8+f12,     f14=f8-f12,     f4=dm(i2,m1),   f1=pm(i9,m14);
f11=f0*f4;
f13=f1*f5,                      f6=f9-f13;	
f9=f0*f5,       f13=f11+f13,                    f11=dm(i4,m1);
f13=f1*f4,      f8=f11+f13,     f10=f11-f13;
	    /*________Do N/4-1 radix-4 butterflies_______*/
lcntr=N/4-1, do lstage until lce;
								f13=f9-f13,     f4=dm(i7,m1),   f5=pm(i15,m9);
				f2=f2+f6,       f15=f2-f6,   	f0=dm(i1,m7),   f1=pm(i9,m14);
f8=f0*f4,       f3=f8+f12,      f7=f8-f12,                      f9=pm(i12,m9);
f12=f1*f5,      f9=f9+f13,      f11=f9-f13,     f13=f2;
f8=f0*f5,       f12=f8+f12,                     f0=dm(i1,m7),   f5=pm(i11,m9);
f13=f1*f4,      f9=f9+f13,      f6=f9-f13,      f4=dm(i3,m1),   f1=pm(i9,m14);
f8=f0*f4,                       f2=f8-f13,      dm(i0,m2)=f3,   pm(i8,m10)=f9;
f13=f1*f5,      f11=f11+f14,    f7=f11-f14,     dm(i0,m2)=f7,   pm(i8,m10)=f6;
f9=f0*f5,       f8=f8+f13,                      f0=dm(i1,m7),   f5=pm(i10,m9);
f13=f1*f4,      f12=f8+f12,     f14=f8-f12,     f4=dm(i2,m1),   f1=pm(i9,m14);
f11=f0*f4,      f3=f10+f15,     f8=f10-f15,                     pm(i8,m10)=f11;
f13=f1*f5,                      f6=f9-f13,      dm(i0,m2)=f3,   pm(i8,m10)=f7;
f9=f0*f5,       f13=f11+f13,                    f11=dm(i4,m1);
lstage:
f13=f1*f4,      f8=f11+f13,     f10=f11-f13,    dm(i0,m2)=f8;

				f13=f9-f13;
	f2=f2+f6,       f15=f2-f6;
	f3=f8+f12,      f7=f8-f12,                      f9=pm(i12,m9);
	f9=f9+f13,      f11=f9-f13,     dm(i0,m2)=f3;
	f9=f9+f2,       f6=f9-f2,       dm(i0,m2)=f7;
									pm(i8,m10)=f9;
	f11=f11+f14,    f7=f11-f14,                     pm(i8,m10)=f6;
	f3=f10+f15,     f8=f10-f15,                     pm(i8,m10)=f11;
									dm(i0,m2)=f3,   pm(i8,m10)=f7;

	rts(db);
	dm(i0,m2)=f8;
	bit clr mode1 BR0 | BR8 | CBUFEN | PEYEN | BDCST1 | BDCST9; 

